__all__ = ['zzInteraction']

from numpy import array
from qutip import qeye, destroy, tensor

def zzInteraction(Qblist, interaction_mat):
    """Creates ZZ-interaction term for hamiltonian
    Inputs:
    - Qblist = list of Qubit objects
    - interaction_mat = symmetric matrix containing the coupling constants between qubits as off diagonal elements
    Outputs: Constant zz-interaction term for the Hamiltonian"""
    eye_vec = []
    H_intlist = []
    H_interaction = 0
    inter = 0

    if type(interaction_mat) == list:
        interaction_mat = array(interaction_mat)
    for QB in enumerate(Qblist):  #creates array of identity matrices of correct dimensions
        eye_vec.append(qeye(QB[1].level))

    for qb in enumerate(Qblist): #enumerates through array of Qubits and creates interaction operators qubits that interacts
        eyeqb = eye_vec.copy()
        eyequb = eye_vec.copy()

        for qub in enumerate(Qblist):
            if interaction_mat[qb[0], qub[0]] > 0 and qb[0] != qub[0]:
                eyeqb[qb[0]] = destroy(qb[1].level).dag() * destroy(qb[1].level)
                eyequb[qub[0]] = destroy(qub[1].level).dag() * destroy(qub[1].level)

                inter = inter + (interaction_mat[qb[0],qub[0]]/2) * tensor(eyeqb)*tensor(eyequb)
                H_intlist.append(inter)

    for ind in range(0, len(H_intlist)): #adds interaction terms together and returns hamiltonian term for ZZ-interaction
        H_interaction = H_interaction + H_intlist[ind]

        ind += ind

    return H_interaction





__all__ = ['anharmonicity', 'AnHarm']

from qutip import create, destroy, qeye, tensor

def anharmonicity(Qblist):
    """function for creating anharmonicty term for the Hamiltonian.
    Qblist[i].Anharm = U (anharmonicity freq) for the i:th qubit
    'Anharm' function tensors the gate out correctly
    Input:
    -Qblist = list of Quubit objects
    Output: Anharmonicity term for the Hamiltonian"""
    H = 0
    for i in range(len(Qblist)):
        H = H + Qblist[i].anharm/2*AnHarm(Qblist, i)
    return H

def AnHarm(Qblist, target):
    """Creates anharmonicity term of correct dimension
    Input is list of qubits and which qubit you want to target with the operator
    Returns Qobj anharmonicity operator for the targeted qubit
    Input:
    - Qblist = list of Qubit objects
    - target = qubit to be targeted by the gate
    Output: anharmonicity term for one of the qubits with correct dimensions"""
    AH = [qeye(Qb.level) for Qb in Qblist]
    AH[target] = create(Qblist[target].level)*create(Qblist[target].level)*destroy(Qblist[target].level)*destroy(Qblist[target].level)
    return tensor(AH)



__all__ = ['envelopeFunc', 'timeFunc']

from numpy import sin, pi, heaviside


def envelopeFunc (t, beta, t_d, t_st):
    """ Models the drive pulse as E(t) = A*sin^2(t*π/t_d),
     Input: -beta = drive strength
            -t_d = drive time
            -t_st = start time (not used since we use different tlist for each step, => t_st = 0 always)
    Output: E(t) = A*sin^2(t*π/t_d)
    """
    E  = beta*sin((t-t_st)*pi/t_d)**2
    return E*heaviside(t_st+t_d-t,1)*heaviside(t-t_st,1)


def timeFunc (t, args):
    """
    Used for creating QobjEvos in the GateFuncs.TimeDepend function, which specifies the args for a given gate.
    Calculates the drive strength (beta) from the max time ~ π rotation (2π for 2qb gate) and
    the drive time as being a an angular fraction of the max time.
    Input: - Drive angle
           - Max gate time (~ ang=π 1qb gates , 2π CZ)
           - Start time for drive (always =0 in current solution)
           - t = element of tlist
    Output: Envelope function E(t)
    """
    ang  = args[0]  # Drive angle 
    t_m  = args[1]  # Max gate time (~ ang=π 1qb gates , 2π CZ)
    t_st = args[2]  # Start time for drive 
    if t_m < 100*1e-9:   # 2qb gates have longer drive time than 100ns
        beta = pi/t_m    # Drive strength corresponds to π drive angle for 1 qb gates
        t_d = t_m * ang / pi  # Drive time for specified angle
    else:
        beta = 2*pi/t_m  # Drive strength should correspond to 2π drive angle for 2qb gates
        t_d = t_m*ang/(2*pi)
    return envelopeFunc(t, beta, t_d, t_st)




__all__ = ['AlgStep', 'createGatesFromStep', 'timeDepend']

"""
Contains: 
- Class Add_step for initialising a new step in the algorithm. 
- Function CreateHFromStep(step, Qblist) that returns a list of [H_real, H_virt, tmax], where
  H_real is the Hamiltonian (Qobj) for the real part of the step, H_virt is the Hamiltonian for the virtual
  part of the step (Qobj), and tmax is the maximal time corresponding to a rotation of π for 1qb gates and 2π for CZ.
- Function TimeDepend(step, gates, t_max) which returns [H,tlist], H being a time dependent Hamiltonian (QobjEvo)
  and tlist being a list of times over which the simulation will run.
"""

from . import gateLib
from qutip import QobjEvo
from numpy import pi, linspace, zeros
from . import qubitClass as Qb
from .envelopeFunction import *
from sys import exit
from . import AnHarm

class AlgStep:
    """
    Class for creating each step in an algorithm.
    Initialises name of gate, target qubit(s) (Tar_Con) and angle of rotation
    """
    def __init__(self, name, Tar_Con, angle):
        self.name = name
        self.Tar_Con = Tar_Con
        self.angle = angle


def createGatesFromStep(step, Qblist, t_max):
    """ Create two lists of Qobj from a step in the step_list, one for virtual and one for real gates
    Also return the max drive time depending on what gates there are in the step
    t_max[0] = max for 1qb gate (~20ns) and t_max[1] = max for 2ab gate (~200ns)
    Input:
    -step = AlgStep object
    -Qblist = list of Qubit objects
    -t_max = list of max gate times for 1 and 2qb gates
    Output: list of real gates, list of virtual gates and max time for the step"""

    anyPhysicalGate = False  # assume no physical gates until they're detected.
    H_real = []
    H_virt = []
    tmax = t_max[0]  # Defaults to time for single qubit gate
    if len(step.name) > len(Qblist):
        print('Error: More gates than qubits have been put to a single depth.')
        exit(1)  # Stops the program with an error code stating that it did not run as it should
    for i in range(len(step.name)):
        try: 
            y = eval("gateLib." + step.name[i])  # Calls the gate corresponding to the step.name[i]
        except Exception as error:
            print('Error: A gate you are trying to perform cannot be executed. \
            \nQNAS only handles gates avaliable at Chalmers quantum computer')
            raise exit(1) # Stops the program
        if type(step.Tar_Con[i]) == list and max(step.Tar_Con[i]) > len(Qblist) - 1 or type(step.Tar_Con[i]) == int and \
                step.Tar_Con[i] > len(Qblist) - 1:
            print('Error: Qubit outside of the number of qubits is being targeted by Tar_Con')
            exit(1)  # Stops the program with the same error code as above
        if step.angle[i] < 0 and not (gateLib.isVirtual(step, i) or step.name[i] in ["CZ", "HD"]):
            print("Warning! Negative angle of " + str(round((step.angle[i]/pi),3)) +'π detected,' + " will be converted to " + str(round((step.angle[i] % (2*pi))/pi,3)) + "π")
            step.angle[i]=step.angle[i] % (2*pi)
        if step.angle[i] > 2 * pi and not (gateLib.isVirtual(step, i) or step.name[i] in ["CZ", "HD"]):
            print("Warning! HUGE angle of " + str(round((step.angle[i]/pi),3)) +'π detected,' + " will be converted to " + str(round((step.angle[i] % (2*pi))/pi,3)) + "π")
            step.angle[i]=step.angle[i] % (2*pi)
        if gateLib.isVirtual(step, i):
            H_virt.append(y(Qblist, step.Tar_Con[i], step.angle[i]))
        elif gateLib.isPhysicalGate(step, i):
            anyPhysicalGate = True
            H_real.append(y(Qblist, step.Tar_Con[i]))
        elif gateLib.isTwoQubitGate(step, i):
            anyPhysicalGate = True
            H_real.append(y(Qblist, step.Tar_Con[i]))
            if step.name[i] in ["CZ"]: #this is needed atm
                step.angle[i] = 2* pi
                tmax = t_max[1]
                if Qblist[step.Tar_Con[i][0]].level != 3 or Qblist[step.Tar_Con[i][1]].level != 3:
                    print('Error: CZ can only act on 3 level qubits. Qubit ' + str(step.Tar_Con[i][0]) + " = " + str(Qblist[step.Tar_Con[i][0]].level) + ", Qubit " + str(step.Tar_Con[i][1]) + " = " + str(Qblist[step.Tar_Con[i][1]].level))
                    exit(1)  # Stops the program with the same error code as above
        elif step.name[i] in ["HD"]: # HD gate feels pretty unique so I left it as it was when I found it
            anyPhysicalGate = True
            step.angle[i] = pi/2
            H = gateLib.HD(Qblist, step.Tar_Con[i])
            H_real.append(H[0])
            H_virt.append(H[1])
        else:
            print(f"No gate added for step {step.name}")
    if not anyPhysicalGate:
        tmax = 0
    return H_real, H_virt, tmax


def timeDepend(step, gates, t_max, Qblist):
    """
    Translates the H_real from Qobj to QobjEvo by including the drive envelope corresponding to the
    angle specified in the step, and sums all the QobjEvos together to return the entire Hamiltonian for
    the real part of the step, as well as a time list (tlist) .
    input:
    - step = AlgStep object
    - gates = real gates from createGatesFromStep
    - t_max = max gate time for the step from createGatesFromStep
    - Oblist = list of Qubit objects
    Output: Time dependant Hamiltonian and tlist for the step
    """
    angles = step.angle
    angles = [angles[i] for i in range(len(angles)) if not gateLib.isVirtual(step,i)] # this eliminates the virtual angles
    # Find max drive time for 1qb gates ~ largest drive angle
    if t_max < 100*1e-9:  # Two qubit gates have longer drive time than 100ns
        if angles != []: # If only virtual gates no angles
            t_dmax = t_max * abs(max(angles)) / pi  # Drive time for the largest angle in step
            tlist = linspace(0, t_dmax, 10)
        else:
            tlist = linspace(0,0,10)
    else:
        tlist = linspace(0,t_max,10)
    args=zeros(3)
    #Create time dep H from angles
    tol = pi/180  # Tolerance for how small angle we can handle, when an angle is "0"
                     # Now set to be able to handle at least one degree and upwards
    H=0
    for i in range(len(step.name)):
        if step.name[i] in ['CZ']:
            for j in range(2): # Removing anharmonicity for the gates targeted by CZ
                target = step.Tar_Con[i][j]
                H = H - Qblist[target].anharm/2 * AnHarm(Qblist, target)
    for i in range(len(gates)):
        if abs(angles[i]) >= tol:  # Can't add gates which have a too small angle
            gate = gates[i]
            args[0] = angles[i]  # Drive angle
            args[1] = t_max  # Theoretical max gate time (~ ang=π)
            args[2] = 0   # Start time for drive
            H = H + QobjEvo([[gate, timeFunc(tlist, args)]], tlist=tlist)
    return H, tlist

if __name__ == "__main__":
    Qblist = []
    Qblist.append(Qb.Qubit(2, [], [], []))
    Qblist.append(Qb.Qubit(2, [], [], []))
    Qblist.append(Qb.Qubit(3, [], [], []))
    Qblist.append(Qb.Qubit(3, [], [], []))

    steps = []
    steps.append(AlgStep(["iswap", "PY", "VPZ"], [[0, 1], 1, 1], [0, pi, pi]))
    hej_real, hej_virt, tlist = createGatesFromStep(steps[0], Qblist, t_max= [20e-9, 20e-9, 20e-9])
    print(hej_real)
    print(tlist)
"""
Ex of usage:

steps = []
steps.append(Add_step(name=["PX", "PX"], Tar_Con = [0,1], angle = [np.pi, 0]))

H = CreateHFromSteps(steps[0],2,2)

"""


__all__ = ['help']

def help():
    """
    User calls this function to recieve information about how to use QnAS
    """

    print("\n\nHello, welcome to QnAS - Quantum noisy Algorithm simulator!\n"
          "Here comes some help on how to get started :)\n\n"
          "The program is made to be able to simulate quantum algorithms, with noise,"
          " from the gates available at WACQT. All gates available are:\n"
          "\"PX\" Pauli x gate\n"
          "\"PY\" Pauli y gate\n"
          "\"VPZ\" Pauli z gate but this one is virtual, takes no time, hence the V in in VPZ\n"
          "\"HD\" Hadamard gate\n"
          "\"CZ\" Controlled z gate\n"
          "\"iSWAP\" iSWAP gate - Available in the iSWAP DLC, coming soon!\n"
          "Make sure these are the only gates you have included in your circuit to create the algorithm!\n"
          "\nDescription of functions: \n\n"
          "help()\n"
          "\t:You already called this function so you probably know that "
          "it just helps you get started!\n\n"
          "solve(Qbfile, circuit, zz_int, ntraj, tmax, store_time_dynamics, e_ops)\n"
          "\t:param Qbfile: Filepath that holds qubit parameters. Default - 3 levels, No noises, anharmonicity -225e6*2*pi [optional]\n"
          "\t:param circuit: Filepath that specifies circuit file.\n"
          "\t:param zz_int: symmetrical square (n x n) matrix that describes interaction between qubits. [optional]\n"
          "\t:param ntraj: number of trajectories for the Monte Carlo solver. Default - 500 [optional]\n"
          "\t:param tmax: Max time for 1qb-gate and 2qb-gate ~ [t_1qb, t_2qb]. Default - [20e-9, 200e-9] [optional]\n"
          "\t:param store_time_dynamics: True/False value to store time dynamics. Default - False [optional]\n"
          "\t:param e_ops: Expectation value operators for store_time_dynamics. Given as [[e_op1, Tar_Con],[e_op2, Tar_Con], ...] [optional]\n"
          "\t:return: if store_time_dynamics is True: Returns ntraj many final states and two lists with exp values "
          "and corresponding times. Else: ntraj many final states\n\n"
          "The qubit parameter file needs to be located in the folder where you are running your program.\n"
          "Here is an example qubit file structure for 3 qubits (OBS! .csv file!):\n"
          "If you want more qubits, just add similar rows to the file!\n\n"
          "\tQubit;Relaxation;Dephasing;Thermal;Anharmonicity(MHz);Levels\n"
          "\t01;1.1;1.2;0.00;-229;3\n"
          "\t02;1.3;1.0;0.00;-225;3\n"
          "\t03;1.2;1.4;0.00;-226;3\n\n"
          "Here comes an example of how you can use qnas.solve() to solve a very simple system of 3\n"
          "qubits given that you have a qubit parameter file called 'qubitData.csv':\n\n"
          "\timport qnas\n"
          "\timport qiskit\n"
          "\tfrom numpy import pi\n"
          "\tcircuit = qiskit.QuantumCircuit(3)\n"
          "\tcircuit.rx(pi,0)\n"
          "\tcircuit.rx(pi,1)\n"
          "\tcircuit.rx(pi,2)\n"
          "\tfinalStates = qnas.solve(Qbfile='qubitData.csv', circuit=circuit)\n\n"
          "If you want descriptions about other functions that you want to use,\n"
          "Go read in the source descriptions under the function definitions!\n"
          "https://github.com/isakwi/Kandidatarbete/tree/main/qnas\n\n"
          "Do you want to donate to QnAS development to keep the project going forward?\n"
          "Swish to +46 70 603 61 27, thanks!")
